#!/usr/bin/env python3
# Copyright 2022, Nice Guy IT, LLC. All rights reserved.
# SPDX-License-Identifier: MIT
# Source: https://github.com/NiceGuyIT/synology_abfb_log_parser
"""
**IMPORTANT**
This script will install the "synology_abfb_log_parser" Python modules in the TRMM Python distribution.
Use at your own risk. Existing modules are not upgraded.

The '--auto-upgrade' parameter will upgrade the "synology_abfb_log_parser" module on every run. This switch is disabled
by default because applications should not auto-update in production unless specifically authorized.
If you're lazy and don't mind an occasional hiccup, create a script check with '--auto-upgrade' and schedule it
to run daily or weekly.

*Note*: When adding arguments to the script in TRMM, use an equals sign "=" to separate the parameter from the value.
For example, use this:
  --ago-unit=hours --ago-value=3
Do not use this as it generate an error:
  --ago-unit hours --ago-value 3


This example will return logs where the "last_backup_status" is *not* "complete". For example:
    --ago-unit=hours --ago-value=12 -> Search the logs for the past 12 hours and return entries
        where the "backup_result.last_backup_status" is not "complete".

$ python3 trmm-example-3.py --help
usage: trmm-example-3.py [-h]
                         [--log-level {debug,info,warning,error,critical}]
                         [--log-path LOG_PATH] [--log-glob LOG_GLOB]
                         [--ago-unit AGO_UNIT] [--ago-value AGO_VALUE]
                         [--auto-upgrade]

Parse the Synology Active Backup for Business logs.

optional arguments:
  -h, --help            show this help message and exit
  --log-level {debug,info,warning,error,critical}
                        set log level for the Synology Active Backup for
                        Business module
  --log-path LOG_PATH   path to the Synology log files
  --log-glob LOG_GLOB   filename glob for the log files
  --ago-unit AGO_UNIT   time span unit, one of [seconds, minutes, hours, days,
                        weeks]
  --ago-value AGO_VALUE
                        time span value
  --auto-upgrade        auto-upgrade the synology_abfb_log_parser module

"""
import argparse
import logging
import pkg_resources
import subprocess
import sys
import traceback


def pip_install_upgrade(modules, logger=logging.getLogger(), upgrade=False):
    """
    Install or upgrade the specified Python modules using 'pip install'.
    :param modules: set of modules to install/upgrade
    :param logger: logging instance of the root logger
    :param upgrade: Bool If True, upgrade the modules.
    :return: None
    """
    if not modules:
        return
    required_modules = set(modules)
    try:
        python = sys.executable
        logger.info(f'Installing/upgrading modules: {required_modules}')
        if upgrade:
            subprocess.check_call([python, '-m', 'pip', 'install', '--upgrade', *required_modules], stdout=subprocess.DEVNULL)
        else:
            subprocess.check_call([python, '-m', 'pip', 'install', *required_modules], stdout=subprocess.DEVNULL)
    except subprocess.CalledProcessError as err:
        logger.error(f'Failed to install/upgrade the required modules: {required_modules}')
        logger.error(err)
        exit(1)


# Check if the modules are installed
try:
    import datetime
    import glob
    import synology_abfb_log_parser
except ModuleNotFoundError:
    required = {'synology_abfb_log_parser'}
    installed = {pkg.key for pkg in pkg_resources.working_set}
    missing = required - installed
    if missing:
        pip_install_upgrade(**{
            'modules': missing
        })
        # Import the modules if they were just installed. Duplicate imports are ignored.
        import datetime
        import glob
        import synology_abfb_log_parser


def main(logger=logging.getLogger(), ago_unit='days', ago_value=1, log_path=None, log_glob='log.txt*'):
    """
    Main program
    :param logger: logging instance of the root logger
    :param ago_unit: string Units of datetime.timedelta. One of ['weeks', 'days', 'hours', 'minutes', 'seconds']
        Note: 'years' and 'months' is not valid.
        See timedelta docs for details: https://docs.python.org/3/library/datetime.html#timedelta-objects
    :param ago_value: int Value of datetime.timedelta
    :param log_path: string Path to the log files.
    :param log_glob: string Filename glob for the log files. Defaults to 'log.txt*'
    :return: None
    """
    after = datetime.timedelta(**{ago_unit: ago_value})

    # Since the package was imported, the syntax is package.subpackage.Class()
    synology = synology_abfb_log_parser.abfb_log_parser.ActiveBackupLogParser(
        # Search logs within the period specified.
        # timedelta() will be off by 1 minute because 1 minute is added to detect if the log entry is last year vs.
        # this year. This should be negligible.
        after=after,

        # Use different log location
        log_path=log_path,

        # Use different filename globbing
        filename_glob=log_glob,

        # Pass the logger
        logger=logger
    )

    # Load the log entries
    if log_path:
        logger.debug(f'Loading log files in "{log_path}"')
    else:
        logger.debug(f'Loading log files in the default location')
    synology.load()

    # Search for entries that match the criteria.
    find = {
        'method_name': 'server-requester.cpp',
        'json': {
            'backup_result': {
                'last_backup_status': {
                    # Find all records with last_backup_status
                }
            }
        },
    }
    logger.debug('Searching the log files')
    found = synology.search(find=find)
    ts = (datetime.datetime.now() - after).strftime('%Y-%m-%d %X')
    if not found:
        logger.info(f"No log entries found since {ts}")
        return

    # True if log entries were found
    errors_found = False

    # Print the log events
    logger.debug('Printing the results')
    print(f'Log entries were found since {ts}:')
    for event in found:

        try:
            # Need to check if the keys are in the event. An error is thrown if a key is accessed that does not exist.
            if 'json' not in event or event['json'] is None:
                continue
            if 'backup_result' not in event['json']:
                continue
            if 'last_success_time' not in event['json']['backup_result']:
                continue
            if 'last_backup_status' not in event['json']['backup_result']:
                continue

            # Nicely formatted timestamp
            ts = event['datetime'].strftime('%Y-%m-%d %X')
            ts_backup = datetime.datetime.fromtimestamp(event['json']['backup_result']['last_success_time'])
            delta_backup = datetime.datetime.now() - ts_backup
            if event['json']['backup_result']['last_backup_status'] != 'complete':
                errors_found = True

                # Always print the output, so it's visible to the users.
                print(f"{ts}: {event['json']['backup_result']}    Days/Hours ago: {delta_backup}")

        except TypeError as err:
            logging.warning(f'Failed to check for key before using. Skipping this event. ERR: {err}')
            logging.warning(traceback.format_exc())
            logging.warning(f'Event: {event}')
            continue

    if errors_found:
        # Errors found. Exit with failure
        exit(1)
    else:
        # No errors found. Exit successful
        exit(0)


# Main entrance here...
if __name__ == '__main__':
    # Parse command line arguments
    parser = argparse.ArgumentParser(description='Parse the Synology Active Backup for Business logs.')
    parser.add_argument('--log-level', default='info', dest='log_level',
                        choices=['debug', 'info', 'warning', 'error', 'critical'],
                        help='set log level for the Synology Active Backup for Business module')
    parser.add_argument('--log-path', default='', type=str,
                        help='path to the Synology log files')
    parser.add_argument('--log-glob', default='log.txt*', type=str,
                        help='filename glob for the log files')
    parser.add_argument('--ago-unit', default='hours', type=str,
                        help='time span unit, one of [seconds, minutes, hours, days, weeks]')
    parser.add_argument('--ago-value', default='1', type=int,
                        help='time span value')
    parser.add_argument('--auto-upgrade', default=False, action='store_true',
                        help='auto-upgrade the synology_abfb_log_parser module')
    args = parser.parse_args()

    # Change default log level to INFO
    default_log_level = 'INFO'
    if args.log_level:
        default_log_level = args.log_level.upper()
    log_format = '%(asctime)s %(funcName)s(%(lineno)d): %(message)s'
    logging.basicConfig(format=log_format, level=default_log_level)
    top_logger = logging.getLogger()
    top_logger.setLevel(default_log_level)

    if args.auto_upgrade:
        requirements = {'synology_abfb_log_parser'}
        pip_install_upgrade(**{
            'modules': requirements,
            'logger': top_logger,
            'upgrade': True,
        })

    main(**{
        'logger': top_logger,
        'log_path': args.log_path,
        'log_glob': args.log_glob,
        'ago_unit': args.ago_unit,
        'ago_value': args.ago_value,
    })



# Synology Active Backup for Business log parser
# Version: 0.1.0
# Author: David Randall
# GitHub: https://github.com/NiceGuyIT/synology_abfb_log_parser
# PyPi: https://pypi.org/project/synology-abfb-log-parser/
# URL: https://NiceGuyIT.biz
#
import datetime
import glob
import json
import logging
import os.path
import re
import sys
import traceback


def fix_single_quotes(json_str):
    """
    fix_single_quotes will replace JSON-type strings containing double quotes with single quotes to make the entire
    string valid JSON.

    Example:
    Given the string
        "task_template": {"backup_cache_content": "{"cached_enabled":false}"}
    the double quotes inside "{...}" will be replaced with single quotes resulting in
        "task_template": {"backup_cache_content": "{'cached_enabled':false}"}

    :param json_str: json_str
    :return cleaned: string
    """
    if not json_str:
        return json_str

    re_left = re.compile(r'"\{')
    re_right = re.compile(r'}"')
    left = re.split(re_left, json_str)
    if not left:
        return json_str

    cleaned = ''
    for index, val in enumerate(left):
        if index == 0:
            # The first value is valid JSON.
            cleaned = val
            continue

        # The right should split into only 2 pieces
        right = re.split(re_right, val)
        if len(right) != 2:
            logging.error('Could not fix JSON with single quotes')
            logging.error(f'JSON string: {json_str}')
            logging.error(f'left: {left}')
            logging.error(f'right: {right}')
            return json_str

        # The first piece is invalid and needs double quotes replaced with single quotes.
        # The second piece is valid JSON
        cleaned += '"{' + right[0].replace('"', "'") + '}"' + right[1]

    return cleaned


def fix_simple(json_str):
    """
    fix_simple will perform simple replacements to fix invalid JSON strings.

    Example:
    Given the string
        "snapshot_info": {"data_length": 18739, }, "subaction": "update_device_spec"
    the ", }" will be replaced with "}" resulting in
        "snapshot_info": {"data_length": 18739}, "subaction": "update_device_spec"

    Given the string
        "volume_name": "\\?\Volume{12345678-1234-abcd-1234-12345678abcd}\"},
    the backslashes are escaped with a backslash resulting in
        "volume_name": "\\\\?\\Volume{12345678-1234-abcd-1234-12345678abcd}\\"},

    :param json_str: json_str
    :return cleaned: string
    """
    if not json_str:
        return json_str
    return json_str.replace(', }', '}').replace('\\', '\\\\')


class ActiveBackupLogParser(object):
    """
    ActiveBackupLogParser will consume Synology Active Backup logs, parse them and make them available for
    searching.
    """

    def __init__(self, after=datetime.timedelta(days=1), log_path=None, filename_glob=None,
                 logger=None):
        """
        Initialize class parameters.

        :param after: datetime.timedelta of how far back to search.
        :param log_path: string path to the log files
        :param filename_glob: string filename glob pattern for the log files
        :param logger: logging instance
        """

        # Logging framework
        if logger is None:
            # If a logger isn't passed, set log level to error.
            self.__logger = logging.getLogger()
            self.__logger.setLevel(logging.ERROR)
        else:
            self.__logger = logger

        # Filename glob for logs
        self.__log_filename_glob = 'log.txt*'
        if filename_glob:
            self.__log_filename_glob = filename_glob

        # Path to log files
        self.__log_path = None
        if sys.platform == 'linux' or sys.platform == 'linux2':
            self.__log_path = '/var/log/activebackupforbusinessagent'
        elif sys.platform == 'darwin':
            self.__log_path = '/var/log/activebackupforbusinessagent'
        elif sys.platform == 'win32':
            self.__log_path = 'C:\\ProgramData\\ActiveBackupForBusinessAgent\\log'
        # Log path was provided
        if log_path:
            self.__log_path = log_path

        # __re_timestamp is a regular expression to extract the timestamp from the beginning of the logs.
        self.__re_log_entry = re.compile(r'^(?P<month>\w{3}) (?P<day>\d+) (?P<time>[\d:]{8}) \[(?P<priority>\w+)] (?P<method_name>[\w.-]+) \((?P<method_num>\d+)\): ?(?P<message>.*)$')

        # __now is a timestamp used to determine if the log entry is after "now". 1 minute is added for
        # processing time.
        self.__now = datetime.datetime.now() + datetime.timedelta(minutes=1)

        # __current_year is the current year and used to determine if the log entry is for this year or last year.
        # The logs do not contain the year.
        self.__current_year = self.__now.year

        # __after is a timestamp used to calculate if the log should be included in the search.
        # Default: 1 day
        if after:
            self.__after = after

        # __events is an array of the log entries.
        self.__events = []

    def load(self):
        """
        Load will load all the log files in the path that have a timestamp after the time delta to search for.

        :return: None
        """
        if not os.path.isdir(self.__log_path):
            self.__logger.error(f'Error: Log directory does not exist: {self.__log_path}')
            return None

        files = glob.glob(os.path.join(self.__log_path, self.__log_filename_glob))
        files.sort(key=os.path.getmtime)
        for file in files:
            if datetime.datetime.fromtimestamp(os.path.getmtime(file)) > datetime.datetime.now() - self.__after:
                self.__logger.debug(f'Processing log file: {file}')
                self.load_log_file(file)

        return None

    def load_log_file(self, log_path):
        """
        load_log_file will iterate over the log files and load the log entries into an object.

        :param log_path: string path to the log files
        :return: None
        """
        # Use the correct encoding.
        # https://stackoverflow.com/questions/17912307/u-ufeff-in-python-string/17912811#17912811
        #   Note that EF BB BF is a UTF-8-encoded BOM. It is not required for UTF-8, but serves only as a
        #   signature (usually on Windows).
        with open(log_path, mode='r', encoding='utf-8-sig') as fh:
            for line in fh.readlines():
                ts_match = self.__re_log_entry.match(line)
                if ts_match:
                    # New log entry
                    # Check if the timestamp is before the threshold
                    # TODO: Use f-strings
                    # ts = datetime.datetime.strptime('{year} {month} {day} {time}'.format(
                    #     month=ts_match.group('month'),
                    #     day=ts_match.group('day'),
                    #     time=ts_match.group('time'),
                    #     year=self.__current_year,
                    # ), '%Y %b %d %X')
                    ts = datetime.datetime.strptime(
                        f'{self.__current_year} {ts_match.group("month")} {ts_match.group("day")} {ts_match.group("time")}',
                        '%Y %b %d %X')
                    if self.__now < ts:
                        # Log timestamp is in the future indicating the log entry is from last year. Subtract one year.
                        # FIXME: This does not take into account leap years. It may be off 1 day on leap years.
                        ts = ts - datetime.timedelta(days=365)

                    if self.__now - self.__after < ts:
                        # Log timestamp is after the 'after' timestamp. Include it.
                        # Always include the timestamp
                        self.__events.append({
                            'datetime': ts,
                            'timestamp': f'{ts_match["month"]} {ts_match["day"]} {ts_match["time"]}',
                            'priority': ts_match['priority'],
                            'method_name': ts_match['method_name'],
                            'method_num': ts_match['method_num'],
                            'message': ts_match['message'].strip(),
                            'json': None,
                        })

                else:
                    # Multiline log entry; append to last line
                    if len(self.__events) == 0:
                        # Log timestamp was before the 'after' window and nothing is captured yet.
                        continue
                    self.__events[len(self.__events) - 1]['message'] += line.strip()

    def parse_json(self, index):
        """
        parse_json will extract the JSON strings from the message and store them in "json".

        :param index: int index of entry to parse
        :return: None
        """
        # Ignore strings that look like JSON but aren't. This is to prevent false JSON parsing errors.
        re_ignore_list = [
            re.compile(r'getVolumeDetailInfo for .*Volume'),
            re.compile(r'Snapshot: \{'),
            re.compile(r'Create snapshot for'),
        ]
        for regex in re_ignore_list:
            matches = re.search(regex, self.__events[index]['message'])
            if matches:
                # Fake JSON found. Don't continue the search.
                self.__logger.debug(f'Ignoring fake JSON: {self.__events[index]["message"]}')
                return

        # If the message has what looks like JSON, extract it from the payload.
        re_list = [
            re.compile(r"'(?P<json>{.*})'"),
            re.compile(r'([^{]*)(?P<json>\{".*})(.*)'),
        ]
        for regex in re_list:
            matches = re.search(regex, self.__events[index]['message'])
            if matches:

                # Fix single quotes
                # Fix commas without values
                json_str = fix_simple(fix_single_quotes(matches['json']))
                try:
                    self.__events[index]['json'] = json.loads(json_str, strict=False)
                    # self.__logger.debug('JSON Object:', self.__events[index]['json'])
                    # Valid JSON found. Don't need to look for more.
                    return
                except json.decoder.JSONDecodeError as err:
                    self.__logger.error('=====')
                    self.__logger.error('ERR: Failed to parse JSON from message')
                    self.__logger.error('Input JSON string:')
                    self.__logger.error(json_str)
                    self.__logger.error('Input log string:')
                    self.__logger.error(self.__events[index]['message'])
                    self.__logger.error(self.__events[index])
                    self.__logger.error(err)
                    self.__logger.error(traceback.format_exc())
                    self.__logger.error('-----')

    def search(self, find):
        """
        search will iterate over the log entries searching for lines that match the values in find.
        find is required.

        :param find: dict representing the log entries to find.
        :return: dict of the log entries.
        """
        for x in range(len(self.__events)):
            self.parse_json(index=x)
            if not self.is_subset(find, self.__events[x]):
                # Event doesn't match search. Mark it for removal.
                self.__events[x] = None
        # Remove events marked for removal.
        self.__events = [x for x in self.__events if x is not None]
        return self.__events

    def is_subset(self, subset, superset):
        """
        is_subset will recursively compare two dictionaries and return true if subset is a subset of the superset.
        See https://stackoverflow.com/a/57675231

        :param subset: dict of the subset
        :param superset: dict of the superset
        :return: true if subset is a subset of the superset
        """
        if subset is None or superset is None:
            return False

        if isinstance(subset, dict):
            return all(key in superset and self.is_subset(val, superset[key]) for key, val in subset.items())

        if isinstance(subset, list) or isinstance(subset, set):
            return all(any(self.is_subset(subitem, superitem) for superitem in superset) for subitem in subset)

        # assume that subset is a plain value if none of the above match
        return subset == superset
